#include "Node.h"
#include "math.h"
#include <sstream>
#include <iomanip>
#include <iostream>
#include <algorithm>
//#include "mex.h"

using namespace std;


// 
// Function that returns node type
//
Node* createNode(int nodeId, EnsembleParams params) {
	if ( params.nodeType == "costsensitive" ) return new NodeCS(0, params);
	if ( params.nodeType == "costsensitivenaive" ) return new NodeCSN(0, params);
	if ( params.nodeType == "classification" ) return new NodeCL(0, params);
	return NULL;
}

Node* createNode(string paramsStr, string testParamsStr, string posteriorStr, EnsembleParams params) {
	if ( params.nodeType == "costsensitive" ) return new NodeCS(paramsStr, testParamsStr, posteriorStr, params);
	if ( params.nodeType == "costsensitivenaive" ) return new NodeCSN(paramsStr, testParamsStr, posteriorStr, params);
	if ( params.nodeType == "classification" ) return new NodeCL(paramsStr, testParamsStr, posteriorStr, params);
	return NULL;
}


//
// Node Base Class
//
Node::Node() {};


vector<float> Node::stringParser(string line) {
	// TODO 
    vector<float> readVals;	
	return readVals;
}

Node::~Node() {

	if (!this->isLeaf) {
		delete left;
		delete right;
	}
}


// for outputting node to file
string Node::nodeToString() {
	stringstream s;
 
    s << nodeId << " | " <<  exsAtNode << " " << isLeaf << " | ";
    for (int cIt = 0; cIt < params.noClasses; cIt++)
		s << fixed << setprecision(5) << posterior[cIt] << " ";
    //s << "| inf " << infoGain << " imp " << impurity;	         
    
    if (!isLeaf)
		s << "| " << dimId << " " << thresh;
    
	return s.str();
}


void Node::doBagging(Matrix2df& X, Matrix2df& Y, vector < vector <int> >& classCnt) {
	// 0 bagging with replacement
    // 1 bagging without replacement (takes a bag prob)
    // 2 no bagging
    // 3 bagging with replacement with an even amount from each class
    
    if (params.baggingType == 0) {
    
        for (int n = 0; n < X.rows; n ++) {
            int randDim = rand()%(X.rows); 
            validExs.push_back(randDim);
        }
    } else if (params.baggingType == 1) {
    
        for (int n = 0; n < X.rows; n ++) {
            if (((float)rand() / (float)RAND_MAX) <= params.bagProb) {
                validExs.push_back(n);
            }
        }
    } else if (params.baggingType == 2) {
        for (int n = 0; n < X.rows; n ++) {
            validExs.push_back(n);
        }
    } else if (params.baggingType == 3) {
        
        // find the class with the minimum num exs
        // WARING there will be a problem if there are no entries for 1 of the classes
        int smallestNumExsInClass = INT_MAX;
        for (int cIt = 0; cIt < params.noClasses; cIt++) {
            if (classCnt[cIt].size() < smallestNumExsInClass)
                smallestNumExsInClass = classCnt[cIt].size();
        }
                
        // sample evenly from each class with replacement
        for (int cIt = 0; cIt < params.noClasses; cIt++) {
            for (int exId = 0; exId < smallestNumExsInClass; exId++) {
                int randDim = rand()%(classCnt[cIt].size());
                validExs.push_back(classCnt[cIt][randDim]);
            }
        }
    }
    
    
    exsAtNode = (float)validExs.size();
	validExsDist.resize((int)exsAtNode);
    
    // sort the indices so they are in ascending order. 
    sort(validExs.begin(), validExs.end());
    
    //mexPrintf("%d %f %d\n", X.rows, exsAtNode, Y.cols);
}


void Node::setAsLeaf(bool deleteChildren) {
	isLeaf = true;
	normalizePosterior();
	
    if (deleteChildren) {
        delete left; 
        left = 0; 
        delete right;
        right = 0;
    }
}

double Node::calcNodeInfoGain(Matrix2df& X, Matrix2df& Y) {

	infoGain = -FLT_MAX;
    
    // if there are no examples at a node - dont compute its impurity
    if( left->exsAtNode > 0 && right->exsAtNode > 0) {
        left->calcNodeImpurity(X, Y);        
        right->calcNodeImpurity(X, Y);
    
        infoGain = 0.0;
        infoGain += left->impurity* (double)(left->exsAtNode / (left->exsAtNode + right->exsAtNode));
        infoGain += right->impurity* (double)(right->exsAtNode / (left->exsAtNode + right->exsAtNode));
        infoGain = impurity - infoGain;
        
       // mexPrintf("%f %f %f | %f %f\n", infoGain, left->impurity, right->impurity, left->exsAtNode, right->exsAtNode);
       
    }

	return infoGain;
}


void Node::getFeatureTest() {
    dimId = rand()%params.noDims;
    thresh = 0.0;
}

void Node::computeMinMax(Matrix2df& X, Matrix2df& Y) {
	tMin = X(validExs[0], dimId);
	tMax = X(validExs[0], dimId);

	for (size_t vIt = 0 ; vIt < validExs.size(); vIt++) {
		int exId = validExs[vIt];

		if (X(exId, dimId) < tMin)
			tMin = X(exId, dimId);

		if (X(exId, dimId) > tMax)
			tMax = X(exId, dimId);

		validExsDist[vIt] = X(exId, dimId);
	}

}

void Node::setRandThresh() {
	thresh = tMin + (tMax-tMin) * ((float)rand() / (float)RAND_MAX);
}


int Node::test(Matrix2df& X, int exId) {
	return (int)(X(exId, dimId) > thresh);		
}


void Node::opPosterior(Matrix2df& res, int exId) {
	for (int cIt = 0; cIt < params.noClasses; cIt++)	
		res.data[exId + cIt*res.rows] = posterior[cIt];	
}

void Node::normalizePosterior() {};
Node::Node(int nodeId_, EnsembleParams params){};
Node::Node(string paramsStr, string testParamsStr, string posteriorStr, EnsembleParams params_){};
void Node::initializeRoot(Matrix2df& X, Matrix2df& Y){};
void Node::train(Matrix2df& X, Matrix2df& Y){};
void Node::updateLeftRight(Matrix2df& X, Matrix2df& Y){};
void Node::calcNodeImpurity(Matrix2df& X, Matrix2df& Y){};
void Node::createChildren(){};
void Node::clearNode(){};

