//TODO clean up, set private - public etc

#ifndef NODE_H
#define NODE_H

#include <vector>
#include "utils.h"

class Node{


public:

	int nodeId;
	float exsAtNode;
	double infoGain;
	double impurity;
	bool isLeaf;
	EnsembleParams params;

	// store the indices of the data pts that land at this node and distances to decision boundary
	std::vector<int> validExs; 
	std::vector<float> validExsDist; 

	std::vector<float> posterior;

	// pointers
	Node *left;   
	Node *right; 
	Node *parent;

	// node test specific	
    int dimId;
    float thresh;
    float tMin;
    float tMax;
		
	// methods that need to be implemented by derived class
	Node(int nodeId_, EnsembleParams params);	
	Node(std::string paramsStr, std::string testParamsStr, std::string posteriorStr, EnsembleParams params_);
	virtual void initializeRoot(Matrix2df& X, Matrix2df& Y);	
	virtual void train(Matrix2df& X, Matrix2df& Y);
	virtual void updateLeftRight(Matrix2df& X, Matrix2df& Y);
	virtual void calcNodeImpurity(Matrix2df& X, Matrix2df& Y);
	virtual void createChildren();
	virtual void clearNode();
	virtual void normalizePosterior();		
	virtual ~Node();	

	// other methods
	Node();
    void getFeatureTest();
    void setRandThresh();
    void computeMinMax(Matrix2df& X, Matrix2df& Y);
	int test(Matrix2df& X, int exId);
	void setAsLeaf(bool deleteChildren);	
	void opPosterior(Matrix2df& res, int exId);	
	std::string nodeToString();	
	std::vector<float> stringParser(std::string line);	
	double calcNodeInfoGain(Matrix2df& X, Matrix2df& Y);
    void doBagging(Matrix2df& X, Matrix2df& Y, std::vector < std::vector <int> >& classCnt);
};

//
// Functions that return a type node depending on parameters sent
//
Node* createNode(int nodeId, EnsembleParams params);
Node* createNode(std::string paramsStr, std::string testParamsStr, std::string posteriorStr, EnsembleParams params);



// Cost Sensitive Node
class NodeCS: public Node {
public:

    std::vector<double> diffM; 
	std::vector<double> diffMSq; 

	NodeCS(int nodeId_, EnsembleParams params_);
	NodeCS(std::string paramsStr, std::string testParamsStr, std::string posteriorStr, EnsembleParams params_);
	~NodeCS();
	void clearNode();
	void createChildren();
	void calcNodeImpurity(Matrix2df& X, Matrix2df& Y);
	void initializeRoot(Matrix2df& X, Matrix2df& Y);
	void train(Matrix2df& X, Matrix2df& Y);
	void updateLeftRight(Matrix2df& X, Matrix2df& Y);
	void normalizePosterior();	
};

// Naive Cost Sensitive Node
class NodeCSN: public Node {
public:
		
	NodeCSN(int nodeId_, EnsembleParams params_);
	NodeCSN(std::string paramsStr, std::string testParamsStr, std::string posteriorStr, EnsembleParams params_);
	~NodeCSN();
	void clearNode();
	void createChildren();
	void calcNodeImpurity(Matrix2df& X, Matrix2df& Y);
	void initializeRoot(Matrix2df& X, Matrix2df& Y);
	void train(Matrix2df& X, Matrix2df& Y);
	void updateLeftRight(Matrix2df& X, Matrix2df& Y);
	void normalizePosterior();	
};


// Classification Node
class NodeCL: public Node {
public:
	NodeCL(int nodeId_, EnsembleParams params_);
	NodeCL(std::string paramsStr, std::string testParamsStr, std::string posteriorStr, EnsembleParams params_);
	~NodeCL();
	void clearNode();
	void createChildren();
	void calcNodeImpurity(Matrix2df& X, Matrix2df& Y);
	void initializeRoot(Matrix2df& X, Matrix2df& Y);
	void train(Matrix2df& X, Matrix2df& Y);
	void updateLeftRight(Matrix2df& X, Matrix2df& Y);
	void normalizePosterior();	
};




#endif