// Classification Node

#include "Node.h"
#include "math.h"
#include <sstream>
//#include "mex.h"

using namespace std;

NodeCL::NodeCL(int nodeId_, EnsembleParams params_) {
	
	isLeaf = false;

	exsAtNode = 0;
	left = 0;
	right = 0;
    parent = 0;
	
	nodeId = nodeId_;
	params = params_;

	posterior.resize(params.noClasses, 0.0);
}

NodeCL::NodeCL(string paramsStr, string testParamsStr, string posteriorStr, EnsembleParams params_) {
    //TODO
}

NodeCL::~NodeCL( ) {
}


// do bagging - set the valid examples at the root
void NodeCL::initializeRoot(Matrix2df& X, Matrix2df& Y) {

    // update posterior
    for (int n = 0; n < exsAtNode; n ++) {
        int curClass = (int)Y(validExs[n], 0);
			posterior[curClass] ++;    
    }
    
	// calc impurity
	calcNodeImpurity(X, Y);

}

// entopy
/*

double log2( double n )  {  
	if(n == 0)
		return(-100000000.0); // hack
	else
		return log( n ) / log( 2.0 );  
}


impurity = 0.0;
for (int cIt = 0; cIt < params.noClasses; cIt++) {
float fract = (float)posterior[cIt]/(float)exsAtNode;
impurity += (fract)*(float)log2(fract);
}
impurity = -impurity;
*/

// expects posterior not to be normalized
void NodeCL::calcNodeImpurity(Matrix2df& X, Matrix2df& Y) {
		
	if (exsAtNode != 0) {
                
        // gini
        impurity = 0.0;
		for (int cIt = 0; cIt < params.noClasses; cIt++) {
			double fract = (double)posterior[cIt]/(double)exsAtNode;
			impurity += (fract)*(fract);
		}
		impurity = 1 - impurity;
	}
	else {
		// no samples at this node
		impurity = FLT_MAX;        
        //impurity = 1 - (1.0/(float)params.noClasses);
	}
}

void NodeCL::train(Matrix2df& X, Matrix2df& Y) {

	left->clearNode();
	right->clearNode();

	for (size_t vIt = 0 ; vIt < validExs.size(); vIt++) {
		int exId = validExs[vIt];
		int curClass = (int)Y(exId, 0); 

        if (validExsDist[vIt] > thresh ) {
        //if (test(X,exId)) {
			right->posterior[curClass] ++;
			right->exsAtNode ++; 
		}
		else {
			left->posterior[curClass] ++;
			left->exsAtNode ++;
		}
	}

}

void NodeCL::clearNode() {
	exsAtNode = 0;
	for (int pIt = 0; pIt < params.noClasses; pIt++)
		posterior[pIt] = 0;
	validExs.clear();
	validExsDist.clear();
}

// updates posteriors and indices
void NodeCL::updateLeftRight(Matrix2df& X, Matrix2df& Y) {

	left->clearNode();
	right->clearNode();

	for (size_t vIt = 0 ; vIt < validExs.size(); vIt++) {
		int exId = validExs[vIt];
		int curClass = (int)Y(exId, 0);

		if (test(X,exId)) {
			right->validExs.push_back(exId);
			right->posterior[curClass] ++;
			right->exsAtNode ++; 
		}
		else {
			left->validExs.push_back(exId);
			left->posterior[curClass] ++;
			left->exsAtNode ++;
		}
	}

    // set the size of the vector of distances for the children to be the same as the vector of valid examples
	left->validExsDist.resize(left->validExs.size());
	right->validExsDist.resize(right->validExs.size());
}


void NodeCL::createChildren(){
	left = new NodeCL(2*(nodeId)+1, params);
	right = new NodeCL(2*(nodeId)+2, params);
			
    left->parent = this;
	right->parent = this;
}

void NodeCL::normalizePosterior() {
	// normalizing posterior so it sums to one
	float denom = posterior[0];
	for (int i=1; i < params.noClasses ; i++) {
		denom = denom + posterior[i];
	}

	for (int i=0; i < params.noClasses ; i++) 
		posterior[i] = posterior[i]/denom;
		
}