// Cost Sensitive Node


#include "Node.h"
#include "math.h"
#include <sstream>
//#include "mex.h"

using namespace std;

NodeCS::NodeCS(int nodeId_, EnsembleParams params_) {

	isLeaf = false;

	exsAtNode = 0;
	left = 0;
	right = 0;	
    parent = 0;

	nodeId = nodeId_;
	params = params_;
		
	posterior.resize(params.noClasses, 0.0);
    
    // for new versions - optimized and unoptimized
    diffM.resize(params.noClasses*params.noClasses - params.noClasses);
    diffMSq.resize(params.noClasses*params.noClasses - params.noClasses);
    
    // for unoptimized version
    //diffM.resize(params.noClasses*params.noClasses);
    //diffMSq.resize(params.noClasses*params.noClasses);
}

NodeCS::NodeCS(string paramsStr, string testParamsStr, string posteriorStr, EnsembleParams params_) {
	//TODO
}

NodeCS::~NodeCS() {
}

void NodeCS::initializeRoot(Matrix2df& X, Matrix2df& Y) {

    // update posterior
    for (int n = 0; n < exsAtNode; n ++) {

		for (int cIt=0; cIt < params.noClasses; cIt ++) {
            posterior[cIt] += Y(validExs[n], cIt);
        }
    
    }
    
	// calc impurity
	calcNodeImpurity(X, Y);
}

// optimized version
void NodeCS::calcNodeImpurity(Matrix2df& X, Matrix2df& Y) {

	if (exsAtNode == 0) {
        // no samples at this node
        impurity = FLT_MAX;
    } else {
        
        impurity = 0.0;
        
        if((nodeId!=0) && (nodeId%2 == 0)) {
            for (int d=0; d < diffM.size(); d++) {  
                diffM[d] = static_cast<NodeCS*>(parent)->diffM[d] - static_cast<NodeCS*>(parent->left)->diffM[d];
                diffMSq[d] = static_cast<NodeCS*>(parent)->diffMSq[d] - static_cast<NodeCS*>(parent->left)->diffMSq[d];
            }
        }
        else {  
        
            for (int i=0; i < diffM.size(); i++) {
                diffM[i] = 0.0;
                diffMSq[i] = 0.0;
            }        
            
            // important that we loop until exsAtNode and not validExs.size(), as there are redundant examples in the vector        
            for (size_t vIt = 0 ; vIt < exsAtNode; vIt++) {
                int exId = validExs[vIt];
                        
                for (int d=0; d < (params.noClasses*params.noClasses - params.noClasses); d++) {            
                
                    // will only be adding a 0 if y_i <= y_j
                    double diffVal = Y.diffMatrix(exId, d);
                    diffM[d] += diffVal;
                    diffMSq[d] += diffVal*diffVal;                
                }
            } 
            
        }

       // just indexing upper triangular matrix
       for (int i=0; i< params.noClasses;i++) {
			for (int j=i+1; j< params.noClasses;j++) {		
                
                // not sure if the indexing of this is correct   
                int ind1 = i+j*params.noClasses - (j + (j<i)); 
                int ind2 = j+i*params.noClasses - (i + (i<j));                 
                double num1 = diffMSq[ind1];
                double num2 = diffMSq[ind2];
                double denom = (diffM[ind1] + diffM[ind2]);
                
                if ( denom != 0.0 ) {
                    double fi1 = num1 / denom;           
                    impurity += (-(fi1*fi1) + fi1);
                        
                    double fi2 = num2 / denom;
                    impurity += (-(fi2*fi2) + fi2);
                }         
			}
		}
               
        // dont want divide by 0
        // if impurity ==0 then leave it at zero  
        // need very small number to prevent weird errors - not sure why
        if (impurity < -0.00000000001 ) {
            impurity = FLT_MAX;
        } else if (impurity > 0.00000000001) {
            impurity = impurity / double((params.noClasses*params.noClasses)-params.noClasses);				
        }
		
	}	
}


/*
// old version
void NodeCS::calcNodeImpurity(Matrix2df& X, Matrix2df& Y) {

	if (exsAtNode == 0) {
        // no samples at this node
        impurity = FLT_MAX;     
    } else {
    
		impurity = 0.0;
        
        // if((nodeId!=0) && (nodeId%2 == 0)) {
            // for (int d=0; d < diffM.size(); d++) {  
                // diffM[d] = static_cast<NodeCS*>(parent)->diffM[d] - static_cast<NodeCS*>(parent->left)->diffM[d];
                // diffMSq[d] = static_cast<NodeCS*>(parent)->diffMSq[d] - static_cast<NodeCS*>(parent->left)->diffMSq[d];
            // }
        // }
        // else {  
            for (int i=0; i < diffM.size(); i++) {
                diffM[i] = 0.0;
                diffMSq[i] = 0.0;
            }
                   
            // important that we loop until exsAtNode and not validExs.size(), as there are redundant examples in the vector
            for (size_t vIt = 0 ; vIt < exsAtNode; vIt++) {
                int exId = validExs[vIt];
                for (int i=0; i< params.noClasses;i++) {
                    for (int j=i+1; j< params.noClasses;j++) {
                        
                        double d_ij = 0.0;
                        double Yi = (double)Y(exId,i);
                        double Yj = (double)Y(exId,j);
                        int ind = 0;
                        
                        if (Yi > Yj) {
                            d_ij = Yi - Yj;
                            ind = i+j*params.noClasses;
                        } else {
                            d_ij = Yj - Yi;
                            ind = j+i*params.noClasses;                    
                        }
                        
                        diffM[ind] += d_ij;
                        diffMSq[ind] += (d_ij*d_ij);
                    }
                }
            }

                
       //}
       
        for (int i=0; i< params.noClasses;i++) {
			for (int j=i+1; j< params.noClasses;j++) {		
                
                // not sure if the indexing of this is correct                    
                double num1 = diffMSq[i+j*params.noClasses];
                double num2 = diffMSq[j+i*params.noClasses];
                double denom = (diffM[i+j*params.noClasses] + diffM[j+i*params.noClasses]);
                
                if ( denom != 0.0 ) {
                    double fi1 = num1 / denom;
                    impurity += (-(fi1*fi1) + fi1);
                        
                    double fi2 = num2 / denom;
                    impurity += (-(fi2*fi2) + fi2);
                }
			}
		}
          
        // dont want divide by 0
        // if impurity ==0 then leave it at zero  
        // need very small number to prevent weird errors - not sure why
        if (impurity < -0.00000000001 ) {
            impurity = FLT_MAX;
        } else if (impurity > 0.00000000001) {
            impurity = impurity / double((params.noClasses*params.noClasses)-params.noClasses);				
        }
	}	
}
*/

void NodeCS::train(Matrix2df& X, Matrix2df& Y) {

    left->exsAtNode = 0;
    right->exsAtNode = 0;
	
	for (size_t vIt = 0 ; vIt < validExs.size(); vIt++) {
		int exId = validExs[vIt];
	
		if (validExsDist[vIt] > thresh ) {
        //if (test(X,exId)) {		
            right->validExs[right->exsAtNode] = exId;
			right->exsAtNode ++; 
		}
		else {	
            left->validExs[left->exsAtNode] = exId;
			left->exsAtNode ++;			
		}
	}
}

// updates posteriors and indices
void NodeCS::updateLeftRight(Matrix2df& X, Matrix2df& Y) {
		
	left->clearNode();
	right->clearNode();

	for (size_t vIt = 0 ; vIt < validExs.size(); vIt++) {
		int exId = validExs[vIt];
		
		if (test(X,exId)) {
			right->validExs.push_back(exId);
			right->exsAtNode ++; 
			for (int cIt=0; cIt < params.noClasses; cIt ++)
				right->posterior[cIt] += Y(exId, cIt);
		}
		else {
			left->validExs.push_back(exId);
			left->exsAtNode ++;
			for (int cIt=0; cIt < params.noClasses; cIt ++)
				left->posterior[cIt] += Y(exId, cIt);
		}
	}
	
	left->validExsDist.resize(left->validExs.size());
	right->validExsDist.resize(right->validExs.size());

}

void NodeCS::clearNode() {
    
	exsAtNode = 0;
	for (int pIt = 0; pIt < params.noClasses; pIt++)
		posterior[pIt] = 0.0;
	validExs.clear();
	validExsDist.clear();
}


void NodeCS::createChildren(){
	left = new NodeCS(2*(nodeId)+1, params);
	right = new NodeCS(2*(nodeId)+2, params);
			
    left->parent = this;
	right->parent = this;
    
    // initialize these to same size as parents
    // we wont need to use all of it
    left->validExs.resize(exsAtNode);
    right->validExs.resize(exsAtNode);
    
}


void NodeCS::normalizePosterior() {
	
	// normalizing posterior so it sums to one
	float denom = posterior[0];
	for (int i=1; i < params.noClasses ; i++) {
		denom += posterior[i];
	}
    
	// need to take into account situation when posterior sums to 0
	if (denom==0.0) {
		for (int i=0; i < params.noClasses ; i++) 
		posterior[i] = (float)(1.0/(float)params.noClasses);
	}
	else {
		for (int i=0; i < params.noClasses ; i++) 
			posterior[i] = posterior[i]/denom;
	}
}