// Naive Cost-Sensitive Node 


#include "Node.h"
#include "math.h"
#include <sstream>
//#include "mex.h"

using namespace std;

NodeCSN::NodeCSN(int nodeId_, EnsembleParams params_) {

	isLeaf = false;

	exsAtNode = 0;
	left = 0;
	right = 0;
    parent = 0;

	nodeId = nodeId_;
	params = params_;

	posterior.resize(params.noClasses, 0.0);
}

NodeCSN::NodeCSN(string paramsStr, string testParamsStr, string posteriorStr, EnsembleParams params_) {
	//TODO
}

NodeCSN::~NodeCSN() {
}

// do bagging - set the valid examples at the root
void NodeCSN::initializeRoot(Matrix2df& X, Matrix2df& Y) {
   
    // update posterior
    for (int n = 0; n < exsAtNode; n ++) {
        for (int cIt=0; cIt < params.noClasses; cIt ++) {
			posterior[cIt] += Y(validExs[n], cIt);
        }
    }
    
    // calc impurity
	calcNodeImpurity(X, Y);
}

void NodeCSN::calcNodeImpurity(Matrix2df& X, Matrix2df& Y) {

	if (exsAtNode != 0) {

        // gini based       
        double sumPost = 0;
        for (int cIt=0; cIt < params.noClasses; cIt ++) {                
            sumPost = sumPost + (double)posterior[cIt];
        }
        
        if (sumPost > 0) {
            impurity = 0.0; 
            for (int cIt=0; cIt < params.noClasses; cIt ++) {
                double normedP = (double)posterior[cIt]/sumPost;
                impurity += normedP*normedP;
            }        
            impurity = 1 - impurity;
        } else {
            // worst impurity possible for gini type
            impurity = 1 - (1.0/(double)params.noClasses);
        }
	}
	else {
		// no samples at this node
		impurity = FLT_MAX;
        //impurity = 1 - (1.0/(float)params.noClasses);
	}
}

void NodeCSN::train(Matrix2df& X, Matrix2df& Y) {

	left->clearNode();
	right->clearNode();
	
	for (size_t vIt = 0 ; vIt < validExs.size(); vIt++) {
		int exId = validExs[vIt];
	
		if (validExsDist[vIt] > thresh ) {
			right->exsAtNode ++; 
			for (int cIt=0; cIt < params.noClasses; cIt ++)
				right->posterior[cIt] += Y(exId, cIt);
		}
		else {
			left->exsAtNode ++;				
			for (int cIt=0; cIt < params.noClasses; cIt ++)
				left->posterior[cIt] += Y(exId, cIt);
		}
	}
}

// updates posteriors and indices
void NodeCSN::updateLeftRight(Matrix2df& X, Matrix2df& Y) {
		
	left->clearNode();
	right->clearNode();

	for (size_t vIt = 0 ; vIt < validExs.size(); vIt++) {
		int exId = validExs[vIt];
		
		if (test(X,exId)) {
			right->validExs.push_back(exId);
			right->exsAtNode ++; 
			for (int cIt=0; cIt < params.noClasses; cIt ++)
				right->posterior[cIt] += Y(exId, cIt);
		}
		else {
			left->validExs.push_back(exId);
			left->exsAtNode ++;
			for (int cIt=0; cIt < params.noClasses; cIt ++)
				left->posterior[cIt] += Y(exId, cIt);
		}
	}
	
	left->validExsDist.resize(left->validExs.size());
	right->validExsDist.resize(right->validExs.size());

}

void NodeCSN::clearNode() {
	
	exsAtNode = 0;
	for (int pIt = 0; pIt < params.noClasses; pIt++)
		posterior[pIt] = 0.0;
	validExs.clear();
	validExsDist.clear();
}


void NodeCSN::createChildren(){
	left = new NodeCSN(2*(nodeId)+1, params);
	right = new NodeCSN(2*(nodeId)+2, params);
			
    left->parent = this;
	right->parent = this;
}


void NodeCSN::normalizePosterior() {
	
	// normalizing posterior so it sums to one
	float denom = posterior[0];
	for (int i=1; i < params.noClasses ; i++) {
		denom += posterior[i];
	}
    
	// need to take into account situation when posterior sums to 0
	if (denom==0.0) {
		for (int i=0; i < params.noClasses ; i++) 
		posterior[i] = (float)(1.0/(float)params.noClasses);
	}
	else {
		for (int i=0; i < params.noClasses ; i++) 
			posterior[i] = posterior[i]/denom;
	}
}