#include "Ensemble.h"
#include "utils.h"
#include <string>

#include "mex.h"
#include "class_handle.h"

using namespace std;

string getString(const mxArray *charAr) {    
    char *input_buf;
    input_buf = mxArrayToString(charAr);
	string charStr(input_buf);        
    mxFree(input_buf);
    
    return charStr;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {	
    
    // Notes
    // indices have to be +2 (or +1) as the string and this object are the first two
    
    // Get the command string
    char cmd[64];
    mxGetString(prhs[0], cmd, sizeof(cmd));

    // Return a handle to a new C++ instance
    if (!strcmp("new", cmd)) {        
        
        // set params
        EnsembleParams params;
        params.noClasses = mxGetScalar(prhs[1]); 
        params.numTrees = mxGetScalar(prhs[2]);
        params.treeDepth = mxGetScalar(prhs[3]);
        params.numRandomTests = mxGetScalar(prhs[4]);
        params.numThreshTests = mxGetScalar(prhs[5]);
        params.baggingType = mxGetScalar(prhs[6]);
        params.bagProb = mxGetScalar(prhs[7]);
        params.minNoOFExsAtNode = mxGetScalar(prhs[8]);        
		params.nodeType = getString(prhs[9]); //costsensitive, costsensitivenaive, classification   
        // warning some business with numclasses for regression versus classification etc
            
    /*
        EnsembleParams params;
        params.noClasses = 2; 
        params.numTrees = 5;
        params.treeDepth = 4;
        params.numRandomTests = 4;
        params.numThreshTests = 30;
        params.baggingType = 0;
        params.bagProb = 0.9;
        params.minNoOFExsAtNode = 1;
        //params.ensembleFile = "op.txt";
		params.nodeType = "classification"; 
      */  
    
        plhs[0] = convertPtr2Mat<Ensemble>(new Ensemble(params)); 
        
        return;
    }
    
    // Delete - destroy the C++ object
    if (!strcmp("delete", cmd)) {      
        destroyObject<Ensemble>(prhs[1]);
        return;
    }
    
    // Get the class instance pointer from the second input
    Ensemble *ensemble_instance = convertMat2Ptr<Ensemble>(prhs[1]);    
    
    // Train    
    if (!strcmp("train", cmd)) {
        // training data
        float *X = (float *)mxGetData(prhs[2]);
        float *Y = (float *)mxGetData(prhs[3]);
        Matrix2df data(mxGetM(prhs[2]), mxGetN(prhs[2]), X);
        Matrix2df labels(mxGetM(prhs[3]), mxGetN(prhs[3]), Y);
       
        ensemble_instance->params.noDims = mxGetN(prhs[2]);
        //mexPrintf("%d\n", ensemble_instance->params.noDims);
        
        // train forest
        ensemble_instance->train(data, labels);
        double *nonLeafNodes;
        plhs[0] = mxCreateDoubleMatrix(1,1, mxREAL);
        nonLeafNodes = mxGetPr(plhs[0]);                
        //*nonLeafNodes = ensemble_instance->countNumNonLeafNodes();
        *nonLeafNodes = ensemble_instance->countNumNodes();
        return;
    }
    
    // Test    
    if (!strcmp("test", cmd)) {                           
        
        float *Xtest = (float *)mxGetData(prhs[2]);
        Matrix2df testData(mxGetM(prhs[2]), mxGetN(prhs[2]), Xtest);
        
        // output 
        mwSize opSize[2];
		opSize[0] = mxGetM(prhs[2]);
		opSize[1] = ensemble_instance->params.noClasses;
        //mexPrintf("%d\n", ensemble_instance->params.treeDepth);
        
        plhs[0] = mxCreateNumericArray(mxGetNumberOfDimensions(prhs[2]), opSize, mxSINGLE_CLASS, mxREAL);
        float *op = (float *)mxGetData(plhs[0]);
        Matrix2df prediction(mxGetM(prhs[2]), ensemble_instance->params.noClasses, op);        
        
        // test forest
        if (nrhs  == 4) {
            // test to specifed depth - 0 is root node, 1 is the first leaf, etc.
            int testDepth = mxGetScalar(prhs[3]); 
            //mexPrintf("depth %d\n", testDepth);
            ensemble_instance->testSpecifyDepth(testData, prediction, testDepth);                    
        } else {
            // else test to full depth
            ensemble_instance->test(testData, prediction);                    
        }
        
        return;
    }
    
    // Write to text file - should be called after train
    if (!strcmp("write", cmd)) {  
        string ensembleFile = getString(prhs[2]);
        ensemble_instance->writeEnsemble(ensembleFile);
    }
}

